/*
 * Copyright (C) 2003 David Sansome <david@dave.sanspic.net>
 */

#include "mutray.h"
#include "pref.h"

#include <qdragobject.h>
#include <kprinter.h>
#include <qpainter.h>
#include <qpaintdevicemetrics.h>

#include <kglobal.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kmenubar.h>
#include <kstatusbar.h>
#include <kkeydialog.h>
#include <kaccel.h>
#include <kconfig.h>
#include <qnetwork.h>
#include <qfile.h>
#include <kmessagebox.h>
#include <qregexp.h>
#include <fcntl.h>
#include <rpm/header.h>
#include <rpm/rpmlib.h>
#include <rpm/rpmdb.h>
#include <kstandarddirs.h>
#include <qtextedit.h>
#include <kimageeffect.h>
#include <kpassivepopup.h>
#include <kpopupmenu.h>


#include <kedittoolbar.h>

#include <kstdaccel.h>
#include <kaction.h>
#include <kstdaction.h>

MUMain::MUMain()
    : KMainWindow( 0, "MUMain" ),
      view(new UpdateList(this))
{
	load();

	srand((unsigned int)time(NULL));
	rpmReadConfigFiles(NULL, NULL);

	updates.setAutoDelete(true);

	KIconLoader iconLoader;
	packageIcon = iconLoader.loadIcon("package",KIcon::Small);
	greyPackageIcon = packageIcon;
	KPixmapEffect::toGray((KPixmap&)greyPackageIcon,true);
	folderIcon = iconLoader.loadIcon("folder",KIcon::Small);
	redFolderIcon = iconLoader.loadIcon("folder_red",KIcon::Small);
	trayIcon = QPixmap(locate("data","mutray/package.png"));
	trayActiveIcon = QPixmap(locate("data","mutray/packagenew.png"));

	tray = new KSystemTray(this);
	tray->setPixmap(trayIcon);
	tray->show();
	iconToggleTimer = new QTimer(this);
	connect(iconToggleTimer,SIGNAL(timeout()),SLOT(toggleIcon()));

	minuteTimer = new QTimer(this);
	connect(minuteTimer,SIGNAL(timeout()),SLOT(aMinuteHasPassed()));
	minuteTimer->start(1000*60);

	// tell the KMainWindow that this is indeed the main widget
	setCentralWidget(view);
	connect(view->list,SIGNAL(currentChanged(QListViewItem*)),SLOT(showDetails(QListViewItem*)));
	connect((QObject*)view->downloadButton,SIGNAL(clicked()),SLOT(download()));
	showDetails(0);

	connect(view->list,SIGNAL(rightButtonClicked(QListViewItem*,const QPoint&,int)),SLOT(popupMenu(QListViewItem*,const QPoint&,int)));

	// then, setup our actions
	setupActions();

	// and a status bar
	statusBar()->insertItem("Idle",0,1);
	statusBar()->show();

	// apply the saved mainwindow settings, if any, and ask the mainwindow
	// to automatically save settings if changed: window size, toolbar
	// position, icon size, etc.
	setAutoSaveSettings();

	retryCount = 0; notFoundCount = 0;
	busy = false;
	if (currentMirror.isEmpty())
		// If no mirror is defined, pick a random one
		updateMirrorList(true);
	else
		checkNow();

	QFile file( "/etc/mandrake-release" );
	file.open(IO_ReadOnly);
	QTextStream stream( &file );
	QRegExp exp(".*release ([\\d.]*) .*");
	exp.setMinimal(true);
	exp.search(stream.readLine());
	distro = exp.cap(1);
}

MUMain::~MUMain()
{
	save();
}

void MUMain::setupActions()
{
    (void)new KAction(i18n("Check for updates now"), 0,
                                  this, SLOT(checkNow()),
                                  actionCollection(), "check_now");
    (void)new KAction(i18n("Download updates"), 0,
                                  this, SLOT(download()),
                                  actionCollection(), "download");

    KStdAction::quit(kapp, SLOT(quit()), actionCollection());

    m_statusbarAction = KStdAction::showStatusbar(this, SLOT(optionsShowStatusbar()), actionCollection());

    KStdAction::keyBindings(this, SLOT(optionsConfigureKeys()), actionCollection());
    KStdAction::preferences(this, SLOT(optionsPreferences()), actionCollection());

    createGUI("mutrayui.rc");
}

void MUMain::save()
{
	KConfig config("mutrayrc");
	config.setGroup("Settings");
	config.writeEntry("Mirror", currentMirror);
	config.writeEntry("AutoCheck", autoCheck);
	config.writeEntry("AutoInterval", autoInterval);
	config.writeEntry("AutoType", autoType);
	config.writeEntry("NotifyOnSecurity", notifyOnSecurity);
	config.writeEntry("NofifyOnBug", notifyOnBug);
	config.writeEntry("NotifyOnNormal", notifyOnNormal);
	config.writeEntry("LastCheck", lastCheck);
	config.writeEntry("IgnoredPackages", ignoredPackages);
}

void MUMain::load()
{
	KConfig config("mutrayrc",true);
	config.setGroup("Settings");
	currentMirror = config.readEntry("Mirror");
	autoCheck = config.readBoolEntry("AutoCheck",true);
	autoInterval = config.readNumEntry("AutoInterval",1);
	autoType = config.readNumEntry("AutoType",1);
	notifyOnSecurity = config.readBoolEntry("NotifyOnSecurity",true);
	notifyOnBug = config.readBoolEntry("NotifyOnBug",false);
	notifyOnNormal = config.readBoolEntry("NotifyOnNormal",false);
	QDateTime* def = new QDateTime(QDate(1970,1,1));
	lastCheck = config.readDateTimeEntry("LastCheck",def); // This seems a very strange way to do it - why does the default have to be a pointer?
	if (*def != lastCheck) delete def;
	ignoredPackages = config.readListEntry("IgnoredPackages");
}

void MUMain::optionsShowStatusbar()
{
    // this is all very cut and paste code for showing/hiding the
    // statusbar
    if (m_statusbarAction->isChecked())
        statusBar()->show();
    else
        statusBar()->hide();
}

void MUMain::optionsConfigureKeys()
{
    KKeyDialog::configureKeys(actionCollection(), "mutrayui.rc");
}

void MUMain::optionsPreferences()
{
    // popup some sort of preference dialog, here
    MUPreferences dlg(this);
    if (dlg.exec())
    {
        // redo your settings
    }
}

void MUMain::aMinuteHasPassed()
{
	if (timeToUpdate()) checkNow();
}

void MUMain::checkNow()
{
	if (busy) return;
	statusBar()->changeItem("Fetching update descriptions...",0);
	busy = true;
	updates.clear();

	// Never, ever use KIO::CopyJob - it doesn't emit a result() signal if the server times out
	// Evil Job... (many hours were spent figuring this out)
	KIO::FileCopyJob* j = new KIO::FileCopyJob(KURL(currentMirror+"/"+distro+"/descriptions"),KURL("file:"+locateLocal("tmp","descriptions")),0644,false,true,false,false);
	connect(j,SIGNAL(result(KIO::Job*)),SLOT(descriptionsFinished(KIO::Job*)));
}


void MUMain::descriptionsFinished( KIO::Job *job )
{
	if (job->error() != 0)
	{
		if (job->error()==11)
			notFoundCount++;
		if (retryCount < 3)
		{
			retryCount++;
			busy=false;
			updateMirrorList(true);  // Try another mirror
			printf("Fetching descriptions failed, trying another mirror (%d)\n",retryCount);
			return;
		}
		if (timeToUpdate())
		{
			lastCheck = QDateTime::currentDateTime();
			save();
		}
		if (notFoundCount>=2)
			KMessageBox::error(this,"The files could not be found on the update mirror.\nMaybe this version of Mandrake Linux is not supported.","Not found");
		else if (isShown())  // We don't want to disturb the user if the window is minimized
			job->showErrorDialog();
		retryCount = 0; notFoundCount = 0;
		busy=false;
		statusBar()->changeItem("Idle",0);
		return;
	}
	//delete job;

	// Parse the file
	updates.clear();
	int section = 0;
	Update* up;
	QFile descriptionsFile(locateLocal("tmp","descriptions"));
	descriptionsFile.open(IO_ReadOnly);
	QTextStream descriptionsStream( &descriptionsFile );
	for (;;)
	{
		QString line = descriptionsStream.readLine();
		if (line.isNull()) break;

		if (line.left(8) == "%package")
		{
			section=0;
			up = new Update;
			up->packagesAffected = QStringList::split(" ",line.right(line.length()-9));
			updates.append(up);
		}
		else if (line.left(4) == "%pre")
			section=1;
		else if (line.left(12) == "%description")
			section=2;
		else
		{
			switch(section)
			{
			case 0:
				if (line.left(8) == "Updated:") up->date = line.right(line.length()-9);
				else if (line.left(11) == "Importance:")
				{
					if (line.right(line.length()-12) == "normal") up->importance = 0;
					else if (line.right(line.length()-12) == "bugfix") up->importance = 1;
					else if (line.right(line.length()-12) == "security") up->importance = 2;
				}
				break;
			case 1: up->reason.append(line+" "); break;
			case 2: up->description.append(line+" "); break;
			}
		}
	}

	KIO::ListJob* j = new KIO::ListJob(KURL(currentMirror+"/"+distro+"/RPMS/"),false);
	statusBar()->changeItem("Fetching package lists...",0);
	connect(j,SIGNAL(entries(KIO::Job*,const KIO::UDSEntryList&)),SLOT(newChildren(KIO::Job*,const KIO::UDSEntryList&)));
	connect(j,SIGNAL(result(KIO::Job*)),SLOT(listFinished(KIO::Job*)));
}

void MUMain::newChildren(KIO::Job*,const KIO::UDSEntryList& infos)
{
	rpmdbOpen(NULL,&dbp,O_RDONLY,0);

	for (uint i=0 ; i<infos.count() ; i++)
	{
		QString fileName;
		for (uint i2=0 ; i2<infos[i].count() ; i2++)
		{
			if (infos[i][i2].m_uds == KIO::UDS_NAME)
			{
				fileName = infos[i][i2].m_str;
				break;
			}
		}
		if (fileName.isEmpty()) continue;
		if (fileName.right(4).lower() != ".rpm") continue;
		QString name = fileName.left(fileName.findRev("-",fileName.findRev("-")-1));

		Header h;
		rpmdbMatchIterator iterator = rpmdbInitIterator(dbp, RPMTAG_NAME, name.latin1(), 0);
		if (iterator)
		{
			h = rpmdbNextIterator(iterator);
			if (h)
			{
				char *version, *release;
				int_32 type, size;
				headerGetEntry(h,RPMTAG_VERSION,&type,(void**)&version,&size);
				headerGetEntry(h,RPMTAG_RELEASE,&type,(void**)&release,&size);
				QString localVersion = QString((const char*)version)+"-"+QString((const char*)release);
				int start = fileName.findRev("-",fileName.findRev("-")-1)+1;
				int end = fileName.findRev(".",fileName.findRev(".")-1);
				QString remoteVersion = fileName.mid(start,end-start);
				printf("%s - %s - %s\n",name.latin1(),remoteVersion.latin1(),localVersion.latin1());
				//if (rpmvercmp(remoteVersion.latin1(),localVersion.latin1()) != 1)
				if (!verCmp(remoteVersion,localVersion))
					removePackage(name);

				headerFree(h);
			}
			else removePackage(name);
		}
		else removePackage(name);
	}
	rpmdbClose(dbp);
}

bool MUMain::verCmp(QString a, QString b)  // Returns true if a is newer than b
{
	// Rpm's rpmvercmp() is really broken...

	if (a==b) return false;

	QStringList aParts = QStringList::split("-",a);
	QStringList bParts = QStringList::split("-",b);
	QStringList aSegments, bSegments;

	for (int part=0 ;  ; part++)
	{
		aSegments = QStringList::split(".",aParts[part]);
		bSegments = QStringList::split(".",bParts[part]);
		for (int i=0 ;  ; i++)
		{
			if (i >= aSegments.count())
			{
				if (i >= bSegments.count())
					break;
				return false;
			}
			if (i >= bSegments.count())
				return true;

			unsigned long aLong, bLong;
			bool aOk, bOk;
			aLong = aSegments[i].toULong(&aOk);
			bLong = bSegments[i].toULong(&bOk);

			if (aOk && bOk)
			{
				if (aLong>bLong) return true;
				else if (bLong>aLong) return false;
			}
			else
			{
				int ret = strcmp(aSegments[i].latin1(),bSegments[i].latin1());
				if (ret == -1) return false;
				if (ret == 1) return true;
			}
		}
	}
	return false;
}

void MUMain::removePackage(QString name)
{
	//printf("%s REMOVED\n",name.latin1());
	for (uint i=0 ; i<updates.count() ; i++)
	{
		QStringList::Iterator it = updates.at(i)->packagesAffected.find(name);
		if (it != updates.at(i)->packagesAffected.end())
		{
			updates.at(i)->packagesAffected.erase(it);
			if (updates.at(i)->packagesAffected.count() == 0)
				updates.remove(i--);
		}
	}
}

void MUMain::listFinished(KIO::Job* job)
{
	statusBar()->changeItem("Idle",0);
	busy=false;
	if (job->error() != 0)
	{
		if (retryCount < 3)
		{
			retryCount++;
			busy=false;
			updateMirrorList(true);  // Try another mirror
			return;
		}
		if (timeToUpdate())
		{
			lastCheck = QDateTime::currentDateTime();
			save();
		}
		retryCount = 0; notFoundCount = 0;
		if (isShown())  // We don't want to disturb the user if the window is minimized
			job->showErrorDialog();
		return;
	}
	//delete job;

	bool notify=false;
	view->list->clear();
	for (uint i=0 ; i<updates.count() ; i++)
	{
		for (uint i2=0 ; i2<updates.at(i)->packagesAffected.count() ; i2++)
		{
			QListViewItem* existingItem = view->list->findItem(updates.at(i)->packagesAffected[i2],0);
			if (existingItem != 0)
			{
				if (importance(existingItem->text(0)) < updates.at(i)->importance)
					delete existingItem;
				else
					continue;
			}
			if ((updates.at(i)->importance == 2) && (notifyOnSecurity)) notify=true;
			if ((updates.at(i)->importance == 1) && (notifyOnBug)) notify=true;
			if ((updates.at(i)->importance == 0) && (notifyOnNormal)) notify=true;
			QListViewItem* parent = view->list->findItem(importance(updates.at(i)->importance),0);
			if (parent == 0)
			{
				parent = new QListViewItem(view->list, importance(updates.at(i)->importance));
				if (updates.at(i)->importance == 2) parent->setPixmap(0,redFolderIcon);
				else parent->setPixmap(0,folderIcon);
				parent->setOpen(true);
			}
			parent = new QListViewItem(parent, updates.at(i)->packagesAffected[i2]);
			if (ignoredPackages.find(updates.at(i)->packagesAffected[i2]) == ignoredPackages.end())
				parent->setPixmap(0,packageIcon);
			else
				parent->setPixmap(0,greyPackageIcon);
		}
	}

	if ((notify) && (timeToUpdate()))
	{
		if (!isShown()) iconToggleTimer->start(1000);
		KPassivePopup::message("MandrakeUpdate","New updates are available",trayActiveIcon,tray);
	}

	if (view->list->childCount() == 0)
	{
		showDetails(NULL,true);
	}

	if (timeToUpdate())
	{
		lastCheck = QDateTime::currentDateTime();
		save();
	}
	retryCount = 0; notFoundCount = 0;
}

void MUMain::showEvent ( QShowEvent * )
{
	iconToggleTimer->stop();
	tray->setPixmap(trayIcon);
}

void MUMain::toggleIcon()
{
	if (trayIconType == 1)
	{
		tray->setPixmap(trayActiveIcon);
		trayIconType = 0;
	}
	else
	{
		tray->setPixmap(trayIcon);
		trayIconType = 1;
	}
	tray->repaint();
}

void MUMain::download()
{
	//KMessageBox::information(this,"The MandrakeUpdate program will now be started to install the updates.\nHave your root password handy.","MandrakeUpdate","mandrakeupdate");
	//KApplication::kdeinitExec("MandrakeUpdate");
	if (busy) return;
	busy = true;
	statusBar()->changeItem("Installing updates...",0);

	QString path = locate("data","mutray/update.sh");
	if (!path)
	{
		KMessageBox::error(this,"The mutray update.sh script could not be found.\nTry reinstalling mutray.","File not found");
		busy = false;
		statusBar()->changeItem("Idle",0);
		return;
	}

	QStringList args;
	args.append("kdesu");
	args.append("-t");
	args.append("-n");
	args.append("-c");
	args.append(path);
	args.append(currentMirror+"/"+distro+"/RPMS/");

	updateScript = new QProcess(args,this);
	if (!updateScript->start())
	{
		KMessageBox::error(this,"Could not start the updater script","Error");
		busy = false;
		statusBar()->changeItem("Idle",0);
		return;
	}

	connect(updateScript,SIGNAL(processExited()),SLOT(downloadFinished()));
	connect(updateScript,SIGNAL(readyReadStdout()),SLOT(readyReadStdout()));
}

void MUMain::downloadFinished()
{
	busy = false;
	checkNow();
}

void MUMain::readyReadStdout()
{
	for (;;)
	{
		QString line = updateScript->readLineStdout();
		if (line.isNull())
			break;
		printf("OUTPUT: %s\n",line.latin1());
	}
}

void MUMain::dataReceived( const QByteArray & datar, QNetworkOperation * )
{
	data.append(QString(datar));
}

void MUMain::dataTransferProgress(int done, int total, QNetworkOperation * )
{
	progressBar->setTotalSteps(total);
	progressBar->setValue(done);
}

void MUMain::updateMirrorList(bool random)
{
	if (busy) return;
	randomMirror = random;
	statusBar()->changeItem("Finding a mirror...",0);
	busy = true;

	KIO::FileCopyJob* j = new KIO::FileCopyJob(KURL("http://www.linux-mandrake.com/mirrorsfull.list"),KURL("file:"+locateLocal("tmp","mirrorlist")),0644,false,true,false,false);
	connect(j,SIGNAL(result(KIO::Job*)),SLOT(mirrorListFinished(KIO::Job*)));
}


void MUMain::mirrorListFinished( KIO::Job *job )
{
	statusBar()->changeItem("Idle",0);
	busy=false;
	if (job->error() != 0)
	{
		if (isShown())  // We don't want to disturb the user if the window is minimized
			job->showErrorDialog();
	}
	else
	{
		processMirrorList();
		if (randomMirror)
		{
			currentMirror = mirrorList[rand() % mirrorList.count()-1];
			checkNow();
		}
	}
	//delete job;
}

void MUMain::processMirrorList()
{
	mirrorList.clear();

	QFile file( "/etc/mandrake-release" );
	file.open(IO_ReadOnly);
	QTextStream stream( &file );
	QString type = "updatesppc";
	if (stream.readLine().right(4) == "i586") type="updatesi586";

	QFile mirrorsListFile(locateLocal("tmp","mirrorlist"));
	mirrorsListFile.open(IO_ReadOnly);
	QTextStream mirrorsListStream( &mirrorsListFile );
	for (;;)
	{
		QString line = mirrorsListStream.readLine();
		if (line.isNull()) break;

		if (line.left(line.find(":")) != type) continue;
		mirrorList.append(line.right(line.length()-line.find(":")-1));
	}

	emit mirrorListChanged();
}

void MUMain::showDetails(QListViewItem* item, bool uptodate)
{
	if (item == 0)
	{
		QString text;
		text  = "<p><b><font color=darkblue>What is MandrakeUpdate?</font></b></p>";
		text += "<p>MandrakeUpdate downloads updated packages for your system from the internet.  These updates fix bugs, and help keep your system secure.</p>";
		text += "<p><b><font color=darkblue>What do I do now?</font></b></p>";
		text += "<p>This program is set to check for updates automatically every hour (you can change this in the settings menu).  If your are connected to the internet, a list of updates will appear on the left.  Click on one for more details.</p>";
		if (uptodate) text += "<p align=center><font size=\"+2\" color=darkblue><b>Your system is fully up to date!</b></font></p>";
		view->details->setText(text);
		return;
	}

	QPtrList<Update> relevantUpdates;
	for (uint i=0 ; i<updates.count() ; i++)
	{
		if (updates.at(i)->packagesAffected.find(item->text(0)) != updates.at(i)->packagesAffected.end())
			relevantUpdates.append(updates.at(i));
	}

	QString text = "";
	if (relevantUpdates.count()>1)
		text = "<p><b><font color=darkblue>Note: </font></b>There are multiple updates to this package.</p><hr>";

	for (uint i=0 ; i<relevantUpdates.count() ; i++)
	{
		text += "<p><b><font color=darkblue>Date: </font></b>"+relevantUpdates.at(i)->date+"</p>";
		text += "<p><b><font color=darkblue>Reason for update: </font></b>"+relevantUpdates.at(i)->reason+"</p>";
		text += "<p><b><font color=darkblue>Package description: </font></b>"+relevantUpdates.at(i)->description+"</p>";
		if (i != relevantUpdates.count()-1) text += "<hr>";
	}

	view->details->setText(text);
}

void MUMain::popupMenu(QListViewItem* item,const QPoint& point,int)
{
	if (item->parent() == 0)
		return;

	bool ignore = (ignoredPackages.find(item->text(0)) != ignoredPackages.end());

	KPopupMenu* menu = new KPopupMenu(this);
	if (ignore)
		menu->insertTitle(greyPackageIcon,item->text(0));
	else
		menu->insertTitle(packageIcon,item->text(0));

	int id = menu->insertItem("Ignore",this,SLOT(toggleIgnore()));
	menu->setItemChecked(id,ignore);

	menu->popup(point);
}

void MUMain::toggleIgnore()
{
	if (ignoredPackages.find(view->list->selectedItem()->text(0)) == ignoredPackages.end())
	{
		ignoredPackages.append(view->list->selectedItem()->text(0));
		view->list->selectedItem()->setPixmap(0,greyPackageIcon);
	}
	else
	{
		ignoredPackages.remove(ignoredPackages.find(view->list->selectedItem()->text(0)));
		view->list->selectedItem()->setPixmap(0,packageIcon);
	}
}

void MUMain::closeEvent( QCloseEvent *e )
{
	MUApplication *muapp = static_cast<MUApplication*>(kapp);
	if (!muapp) return;
	if(muapp->shuttingDown)
	{
		KMainWindow::closeEvent( e );
		return;
	}

	hide();
	e->ignore();
}

#include "mutray.moc"
